use crate::app_config::AppConfig;
use crate::libvirt_lib_structures::{DomainXML, DomainXMLUuid};
use crate::libvirt_rest_structures::*;
use actix::{Actor, Context, Handler, Message};
use virt::connect::Connect;
use virt::domain::Domain;

pub struct LibVirtActor {
    m: Connect,
}

impl LibVirtActor {
    /// Connect to hypervisor
    pub async fn connect() -> anyhow::Result<Self> {
        let hypervisor_uri = AppConfig::get().hypervisor_uri.as_deref().unwrap_or("");
        log::info!(
            "Will connect to hypvervisor at address '{}'",
            hypervisor_uri
        );
        let conn = Connect::open(hypervisor_uri)?;

        Ok(Self { m: conn })
    }
}

impl Actor for LibVirtActor {
    type Context = Context<Self>;
}

#[derive(Message)]
#[rtype(result = "anyhow::Result<HypervisorInfo>")]
pub struct GetHypervisorInfo;

impl Handler<GetHypervisorInfo> for LibVirtActor {
    type Result = anyhow::Result<HypervisorInfo>;

    fn handle(&mut self, _msg: GetHypervisorInfo, _ctx: &mut Self::Context) -> Self::Result {
        let node = self.m.get_node_info()?;
        Ok(HypervisorInfo {
            r#type: self.m.get_type()?,
            hyp_version: self.m.get_hyp_version()?,
            lib_version: self.m.get_lib_version()?,
            capabilities: self.m.get_capabilities()?,
            free_memory: self.m.get_free_memory()?,
            hostname: self.m.get_hostname()?,
            node: HypervisorNodeInfo {
                cpu_model: node.model,
                memory_size: node.memory,
                number_of_active_cpus: node.cpus,
                cpu_frequency_mhz: node.mhz,
                number_of_numa_cell: node.nodes,
                number_of_cpu_socket_per_node: node.sockets,
                number_of_core_per_sockets: node.cores,
                number_of_threads_per_core: node.threads,
            },
        })
    }
}

#[derive(Message)]
#[rtype(result = "anyhow::Result<DomainXMLUuid>")]
pub struct DefineDomainReq(pub DomainXML);

impl Handler<DefineDomainReq> for LibVirtActor {
    type Result = anyhow::Result<DomainXMLUuid>;

    fn handle(&mut self, msg: DefineDomainReq, _ctx: &mut Self::Context) -> Self::Result {
        let xml = serde_xml_rs::to_string(&msg.0)?;
        log::debug!("Define domain:\n{}", xml);
        let domain = Domain::define_xml(&self.m, &xml)?;
        DomainXMLUuid::parse_from_str(&domain.get_uuid_string()?)
    }
}
