use crate::constants;
use crate::libvirt_lib_structures::{DomainMemoryXML, DomainXML, DomainXMLUuid, OSTypeXML, OSXML};
use crate::libvirt_rest_structures::LibVirtStructError::StructureExtractionError;
use lazy_regex::regex;

#[derive(thiserror::Error, Debug)]
enum LibVirtStructError {
    #[error("StructureExtractionError: {0}")]
    StructureExtractionError(&'static str),
}

#[derive(serde::Serialize)]
pub struct HypervisorInfo {
    pub r#type: String,
    pub hyp_version: u32,
    pub lib_version: u32,
    pub capabilities: String,
    pub free_memory: u64,
    pub hostname: String,
    pub node: HypervisorNodeInfo,
}

#[derive(serde::Serialize)]
pub struct HypervisorNodeInfo {
    pub cpu_model: String,
    /// Memory size in kilobytes
    pub memory_size: u64,
    pub number_of_active_cpus: u32,
    pub cpu_frequency_mhz: u32,
    pub number_of_numa_cell: u32,
    pub number_of_cpu_socket_per_node: u32,
    pub number_of_core_per_sockets: u32,
    pub number_of_threads_per_core: u32,
}

#[derive(serde::Serialize, serde::Deserialize)]
pub enum BootType {
    Legacy,
    UEFI,
    UEFISecureBoot,
}

#[derive(serde::Serialize, serde::Deserialize)]
pub enum VMArchitecture {
    #[serde(rename = "i686")]
    I686,
    #[serde(rename = "x86_64")]
    X86_64,
}

#[derive(serde::Serialize, serde::Deserialize)]
pub struct VMInfo {
    /// VM name (alphanumeric characters only)
    pub name: String,
    pub uuid: Option<DomainXMLUuid>,
    pub genid: Option<DomainXMLUuid>,
    pub title: Option<String>,
    pub description: Option<String>,
    pub boot_type: Option<BootType>,
    pub architecture: VMArchitecture,
    /// VM allocated memory, in megabytes
    pub memory: usize,
}

impl VMInfo {
    /// Turn this VM into a domain
    pub fn to_domain(self) -> anyhow::Result<DomainXML> {
        if !regex!("^[a-zA-Z0-9]+$").is_match(&self.name) {
            return Err(StructureExtractionError("VM name is invalid!").into());
        }

        if let Some(n) = &self.uuid {
            if n.is_valid() {
                return Err(StructureExtractionError("VM UUID is invalid!").into());
            }
        }

        if let Some(n) = &self.genid {
            if n.is_valid() {
                return Err(StructureExtractionError("VM genid is invalid!").into());
            }
        }

        if let Some(n) = &self.title {
            if n.contains('\n') {
                return Err(StructureExtractionError("VM title contain newline char!").into());
            }
        }

        if self.memory < constants::MIN_VM_MEMORY || self.memory > constants::MAX_VM_MEMORY {
            return Err(StructureExtractionError("VM memory is invalid!").into());
        }

        Ok(DomainXML {
            r#type: "kvm".to_string(),
            name: self.name,
            uuid: self.uuid,
            genid: self.genid.map(|i| i.0),
            title: self.title,
            description: self.description,

            os: OSXML {
                r#type: OSTypeXML {
                    arch: match self.architecture {
                        VMArchitecture::I686 => "i686",
                        VMArchitecture::X86_64 => "x86_64",
                    }
                    .to_string(),
                    body: "hvm".to_string(),
                },
            },
            memory: DomainMemoryXML {
                unit: "MB".to_string(),
                memory: self.memory.to_string(),
            },
        })
    }
}
